#
# This file is part of the GROMACS molecular simulation package.
#
# Copyright 2019- The GROMACS Authors
# and the project initiators Erik Lindahl, Berk Hess and David van der Spoel.
# Consult the AUTHORS/COPYING files and https://www.gromacs.org for details.
#
# GROMACS is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# GROMACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with GROMACS; if not, see
# https://www.gnu.org/licenses, or write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
#
# If you want to redistribute modifications to GROMACS, please
# consider that scientific software is very special. Version
# control is crucial - bugs must be traceable. We will be happy to
# consider code for inclusion in the official distribution, but
# derived work must not be called official GROMACS. Details are found
# in the README & COPYING files - if they are missing, get the
# official version at https://www.gromacs.org.
#
# To help us fund GROMACS development, we humbly ask that you cite
# the research papers on the package. Check out https://www.gromacs.org.

"""mdrun operation module

The mdrun operation (in its first draft) conforms to the user-level API, but does
not use the Python Context resource manager. It uses either the legacy 0.0.7
Context or its own Context, also implemented in this module.
"""

__all__ = ["mdrun", "SimulationError"]

import hashlib
import inspect
import os
import typing
import warnings
from contextlib import contextmanager
from enum import Enum
from pathlib import Path
from typing import Union

import gmxapi
import gmxapi.abc
import gmxapi.operation as _op
from gmxapi import exceptions
from gmxapi import logger as root_logger
from gmxapi.runtime import scoped_resources, ResourceAssignment

from . import fileio
from . import workflow
from .abc import ModuleObject

# Initialize module-level logger
logger = root_logger.getChild("mdrun")
logger.info("Importing {}".format(__name__))

# Output in the gmxapi.operation Context.
# TODO: Consider using a single base class for the DataProxy, but have distinct
#  data descriptor behavior (or different descriptor implementations in different
#  subclasses) so that static code inspection can more easily determine the
#  attributes of the data proxies.
_output_descriptors = (
    _op.OutputDataDescriptor("directory", str),
    _op.OutputDataDescriptor("checkpoint", str),
    _op.OutputDataDescriptor("parameters", dict),
    _op.OutputDataDescriptor("stderr", str),
    _op.OutputDataDescriptor("stdout", str),
    _op.OutputDataDescriptor("trajectory", str),
)
_publishing_descriptors = {
    desc._name: gmxapi.operation.Publisher(desc._name, desc._dtype)
    for desc in _output_descriptors
}
_output = _op.OutputCollectionDescription(
    **{descriptor._name: descriptor._dtype for descriptor in _output_descriptors}
)


class FD(Enum):
    """Enumerate the canonical posix file descriptors for stdio.

    This provides a reference for use with subprocesses or called code,
    since pytest or other calling code may have already manipulated the
    file descriptors associated with Python's `sys` module.
    """

    STDOUT = 1
    STDERR = 2
    # Provide a type hint for the *value* property.
    value: int


@contextmanager
def redirect_stdio(fd: FD, path: Union[str, Path, os.PathLike]):
    """Temporarily redirect the indicated file descriptor to the indicated file.

    *path* is opened for (appended) writing for the scope of the context manager.
    The indicated file descriptor is duplicated, then replaced with that of
    the opened *path*. On exiting the context manager, *path* is closed, *fd*
    is restored to its original value, and the duplicate is closed.

    Note:
        `contextlib.redirect_stdout`, etc., only works for Python calls that
        already support indirection because they ultimately use the `sys`
        module attributes.
    """
    fd_backup = os.dup(fd.value)
    try:
        with open(path, "a") as fh:
            yield os.dup2(fh.fileno(), fd.value)
    finally:
        os.dup2(fd_backup, fd.value)
        os.close(fd_backup)


class OutputDataProxy(_op.DataProxyBase, descriptors=_output_descriptors):
    """Implement the 'output' attribute of `mdrun` operations.

    Attributes:
        checkpoint (str): Full path to ``cpt`` file.
        directory (str): Full path to the working directory in which the simulation ran.
        parameters (dict): Dictionary of parameters with which the simulation was run.
        stderr (str): Full path to the text file that captured stderr during the simulation.
        stdout (str): Full path to the text file that captured stdout during the simulation.
        trajectory (str): Full path to trajectory output (corresponding to the ``-o``
            flag, if provided).

    Notes:
        For multi-rank MPI-enabled simulators, *stderr* and *stdout* are reported
        for the root rank only, in line with how |Gromacs| behaves.

    .. versionchanged:: 0.4

        Added *directory* output, replacing an earlier "hidden" *_work_dir* output.

    .. versionadded:: 0.4

        *stderr* and *stdout* provide paths to the captured standard I/O.
        Previously, a lot of output from the underlying library bypassed Python
        and went straight to the standard output and standard error of the
        calling process.

    """


class PublishingDataProxy(_op.DataProxyBase, descriptors=_publishing_descriptors):
    """Manage output resource updates for MDRun operation."""


_output_factory = _op.OutputFactory(
    output_proxy=OutputDataProxy,
    output_description=_output,
    publishing_data_proxy=PublishingDataProxy,
)

# Input in the gmxapi.operation Context for the dispatching runner.
# The default empty dictionary for parameters just means that there are no overrides
# to the parameters already provided in _simulation_input.
_input = _op.InputCollectionDescription(
    [
        (
            "_simulation_input",
            inspect.Parameter(
                "_simulation_input",
                inspect.Parameter.POSITIONAL_OR_KEYWORD,
                annotation=str,
            ),
        ),
        (
            "parameters",
            inspect.Parameter(
                "parameters",
                inspect.Parameter.POSITIONAL_OR_KEYWORD,
                annotation=dict,
                default=dict(),
            ),
        ),
        (
            "runtime_args",
            inspect.Parameter(
                "runtime_args",
                inspect.Parameter.POSITIONAL_OR_KEYWORD,
                annotation=dict,
                default=dict(),
            ),
        ),
    ]
)


def _standard_node_resource_factory(*args, **kwargs) -> _op.DataSourceCollection:
    """Translate Python UI input to the gmxapi.operation node builder inputs."""
    source_collection = _input.bind(*args, **kwargs)
    logger.info("mdrun input bound as source collection {}".format(source_collection))
    return source_collection


class SimulationError(exceptions.Error):
    """Error occurred during simulation."""


class LegacyImplementationSubscription(object):
    """Input type representing a subscription to the 0.0.7 implementation.

    This class provides the input parameter type for the gmxapi.mdrun run time
    resource builder. However, the mdrun operation was never fully ported from
    gmxapi 0.0.x to 0.x.

    The current gmxapi.mdrun implementation is mostly a wrapper for the 0.0.7
    implementation. The actual simulation work is performed in this overloaded
    resource factory.

    The gmxapi.mdrun input resource can be created from the standard data of
    the simulation module. The actual simulation work is performed while
    preparing this input resource so that the gmxapi.mdrun operation itself
    merely publishes the results produced here.

    See Also:
        * :issue:`3145`
        * :issue:`3147`
        * :issue:`4079`

    """

    ensemble_rank: typing.Optional[int] = None
    """Member index of the simulation within an ensemble.
    
    The ensemble communicator rank (of the root simulator rank) determines the
    ensemble_rank of all ranks in a simulator.

    Only the first MPI rank of each Simulator is allowed to talk to the other
    Simulators on the ensemble communicator, but it is convenient to use this
    identifier to group all resources associated with a simulation ensemble member.
    
    *ensemble_rank* is either the ensemble member ID or None.
    """

    workdir: typing.List[str]  # Simulation working directories.
    parameters: typing.List[dict]  # MDP dictionaries.
    runtime_args: typing.List[dict]  # CLI args passed as gmxapi 0.0.7 work params.

    def __init__(self, resource_manager: _op.ResourceManager):
        # Avoid importing C++ extension modules until they are clearly needed.
        from .context import Context as LegacyContext
        import gmxapi._gmxapi as _gmxapi
        from mpi4py.MPI import Comm as mpi4py_Comm

        self._gmxapi = _gmxapi

        assert isinstance(resource_manager, _op.ResourceManager)
        # We use several details of the gmxapi.operation.Context.ResourceManager.
        # These dependencies can evolve into the subscription protocol between Contexts.

        # Configure and run a gmxapi 0.0.7 session.
        # 0. Determine ensemble width.
        # 1. Choose, create/check working directories.
        # 2. Create source TPR.
        # 2.5 Update runtime arguments.
        # 3. Create workspec.
        # 3.5 Add plugin potentials, if any.
        # 4. Run.
        # 5. Collect outputs from context (including plugin outputs) and be ready to publish them.

        # Determine ensemble width
        ensemble_width = resource_manager.ensemble_width

        # Choose working directories
        # TODO: operation working directory naming scheme should be centrally well-defined.
        # Note that workflow.WorkSpec.uid is currently dependent on the input file parameter,
        # so we cannot create the input file path in the working directory based on WorkSpec.uid.
        workdir_list = [
            "{node}_{member}".format(node=resource_manager.operation_id, member=member)
            for member in range(ensemble_width)
        ]
        # Define container references of appropriate type for later MPI calls.
        parameters_dict_list = [{}] * ensemble_width
        runtime_args_list = [{}] * ensemble_width
        tpr_filenames = [""] * ensemble_width

        # This is a reasonable place to start using MPI ensemble implementation details.
        # We will want better abstraction in the future, but it is best if related filesystem
        # accesses occur from the same processes, consistently. Note that we already
        # handle some optimization and dependency reduction when the ensemble size is 1.
        # TODO: multithread and multiprocess alternatives to MPI ensemble management.

        # TODO: Consider the allowable hierarchy of resource allocation and assignment.
        # The way that chained simulations are originally implemented, this code will be
        # reentered due to resource_manager.local_input() below. We end up creating
        # multiple coexisting and overlapping subcommunicators, and only the inherent
        # serialization of the input resolution prevents CPUs from being oversubscribed
        # at run time.
        base_context = gmxapi.runtime.BaseContext.instance()
        # Derive resources for the simulation task directly from the base resources.
        base_comm = base_context.communicator()
        assert isinstance(base_comm, mpi4py_Comm)
        base_rank = base_comm.Get_rank()

        # Get resources for the entire ensemble of tasks (the session).
        # Each simulator gets an MPI communicator for the ranks it can use.
        # Additionally, some ensemble simulations are coupled parallel tasks,
        # requiring one rank each in a session-wide communicator. We want an
        # identifier for each ensemble scope, but to make it useful as an MPI_Split
        # color we convert the operation_id string to a 16-bit integer node_id.
        # We don't currently have a specification for
        # identifying which simulations do or do not require the ensemble communicator,
        # so we just make one available to all simulations.
        node_id = int.from_bytes(
            hashlib.md5(resource_manager.operation_id.encode("utf8")).digest(),
            byteorder="little",
            signed=False,
        ) % int(2**16)
        communicator_requirements = tuple(
            {"ensemble_comm": node_id, "subtask_comm": True}
            for _ in range(ensemble_width)
        )
        requirements = gmxapi.runtime.ResourceRequirements(
            communication=communicator_requirements
        )
        with scoped_resources(
            base_context, requirements=requirements
        ) as session_resources:
            assert isinstance(session_resources, ResourceAssignment)
            # If a rank participates in the work, it will have a non-null task communicator.
            # The root rank of each task communicator will also have an ensemble communicator.
            # The ensemble member ID of each participating process is the ensemble communicator rank
            # of the root rank of its task communicator.
            # However, to avoid unnecessary MPI communication, we have stashed this information in
            # the ResourceAssignment.
            simulator_comm: mpi4py_Comm = session_resources.communicator()
            _current_rank_participates = bool(simulator_comm)
            if _current_rank_participates:
                self.ensemble_rank = session_resources.subtask_id()
                assert self.ensemble_rank < ensemble_width

            source_file, parameters, runtime_args = None, None, None
            # For the interim subscription system between MPI ranks, to synchronize
            # ResourceManager.update_output() calls, `local_input` needs to be called
            # on *all* ranks once for each member (participating rank).
            for ensemble_member in range(ensemble_width):
                with resource_manager.local_input(member=ensemble_member) as input_pack:
                    if ensemble_member == self.ensemble_rank:
                        source_file = input_pack.kwargs["_simulation_input"]
                        parameters = input_pack.kwargs["parameters"]
                        runtime_args = input_pack.kwargs["runtime_args"]
                        # If there are any other key word arguments to process from the
                        # gmxapi.mdrun factory call, do it here.

            if _current_rank_participates:
                # TODO: This should be a richer object that includes at least host information
                #  and preferably the full gmxapi Future interface.
                workdir = os.path.abspath(workdir_list[self.ensemble_rank])

                # TODO: We should really name this file with a useful input-dependent tag.
                tprfile = os.path.join(workdir, "topol.tpr")

                expected_working_files = [tprfile]

                if session_resources.is_subtask_root():
                    if os.path.exists(workdir):
                        if not os.path.isdir(workdir):
                            raise exceptions.ApiError(
                                f"Chosen working directory path exists but is not a directory: {workdir}"
                            )
                        # Confirm that this is a restarted simulation.
                        # It is unspecified by the API, but at least through gmxapi 0.1,
                        # all simulations are initialized with a checkpoint file named state.cpt
                        # (see src/api/cpp/context.cpp)
                        checkpoint_file = runtime_args.get("-cpi", "state.cpt")
                        if not os.path.isabs(checkpoint_file):
                            checkpoint_file = os.path.join(workdir, checkpoint_file)
                        expected_working_files.append(checkpoint_file)

                        for file in expected_working_files:
                            if not os.path.exists(file):
                                logger.error(
                                    f"Expected file {file} not found. gmxapi.mdrun task "
                                    f"{resource_manager.operation_id} is in an unknown state. Aborting."
                                )
                                raise exceptions.ApiError(
                                    f"Cannot determine working directory state: {workdir}"
                                )
                    else:
                        # Build the working directory and input files.
                        os.mkdir(workdir)
                        sim_input = fileio.read_tpr(source_file)
                        # TODO(#3295): insertion point for updated positions and velocities.
                        for key, value in parameters.items():
                            try:
                                sim_input.parameters.set(key=key, value=value)
                            except _gmxapi.Exception as e:
                                raise exceptions.ApiError(
                                    "Bug encountered. Unknown error when trying to set simulation "
                                    "parameter {} to {}".format(key, value)
                                ) from e

                        fileio.write_tpr_file(output=tprfile, input=sim_input)
                        logger.info("Created {} on rank {}".format(tprfile, base_rank))

                # When the workflow is using dynamically generated / temporary
                # directories, there may be path elements that are created
                # differently on different ranks. This should not generally matter
                # to us because filesystem I/O is confined to the root rank of
                # each simulator. However, consistency is important for two reasons.
                # 1. Results (which include absolute paths) should be consistently reported.
                # 2. The gmxapi 0.0.7 representation of ensemble work is decoupled from
                #    the resource assignment, and all participating processes receive a
                #    complete representation that is assumed to be consistent across members.
                # Also, debugging is a lot easier if log messages and local variables are
                # consistent across processes.

                # Gather the actual outputs from the ensemble members,
                # then share to participating processes.
                ensemble_comm = session_resources.ensemble_communicator()
                if ensemble_comm:
                    # We should not assume that abspath expands the same on different MPI ranks.
                    # Let each simulation master rank determine its own working directory, etc,
                    # report to the root rank of the parent communicator, then broadcast
                    # to all interpreter processes for consistent results.
                    workdir_list = ensemble_comm.allgather(workdir)
                    tpr_filenames = ensemble_comm.allgather(tprfile)
                    parameters = fileio.read_tpr(tprfile).parameters.extract()
                    parameters_dict_list = ensemble_comm.allgather(parameters)
                    runtime_args_list = ensemble_comm.allgather(runtime_args)
                    for _data in (
                        workdir_list,
                        tpr_filenames,
                        parameters_dict_list,
                        runtime_args_list,
                    ):
                        assert len(_data) == ensemble_width

                workdir_list = simulator_comm.bcast(workdir_list)
                tpr_filenames = simulator_comm.bcast(tpr_filenames)
                parameters_dict_list = simulator_comm.bcast(parameters_dict_list)
                runtime_args_list = simulator_comm.bcast(runtime_args_list)
                for _data in (
                    workdir_list,
                    tpr_filenames,
                    parameters_dict_list,
                    runtime_args_list,
                ):
                    assert len(_data) == ensemble_width

                logger.debug(
                    "Context rank {} acknowledges working directories {}".format(
                        base_rank, workdir_list
                    )
                )
                workdir = workdir_list[self.ensemble_rank]
                logger.debug(
                    "Operation {}:{} will use {}".format(
                        resource_manager.operation_id, self.ensemble_rank, workdir
                    )
                )
                if hasattr(resource_manager, "mdrun_kwargs"):
                    warnings.warn(
                        DeprecationWarning(
                            "Ignoring ResourceManager.mdrun_kwargs attribute. "
                            "Provide runtime arguments to mdrun with the *runtime_args* kwarg."
                        )
                    )
                # TODO(#3718): Normalize the way we pass run time parameters to mdrun.
                kwargs = runtime_args_list[self.ensemble_rank].copy()
                for key, value in runtime_args.items():
                    logger.debug(
                        "Adding mdrun run time argument from user input: {}".format(
                            key + "=" + str(value)
                        )
                    )
                # Note that this violates the traditional gmxapi assumption that all ranks see the same
                # instructions. The "md_sim" element of the gmxapi 0.0.7 workspec ends up being unique
                # to the rank that sees it.
                work = workflow.from_tpr(tpr_filenames, **kwargs)
                self.workspec = work.workspec
                # TODO(#3145): Attach extension code, if any.

                # Go ahead and execute immediately. No need for lazy initialization in this basic case.
                context = LegacyContext(
                    work=self.workspec,
                    workdir_list=workdir_list,
                    resources=session_resources,
                )
                self.simulation_module_context = context
                # Note: The redirection of stdout and stderr here is a workaround (#4541) for
                # inflexible output handling in libgromacs. A better solution would be to use a
                # logging facility instead of assuming terminal I/O in the core library, or at
                # least to set and use file descriptors managed through the output_env or program_context,
                # but, as of resolution of #4541, there are no near term plans to make such changes.
                # See also #1505, #2585, #2999, #3015, #3035
                outfile = "stdout.txt"
                errfile = "stderr.txt"
                if session_resources.subtask_rank() != 0:
                    outfile = f"stdout_rank{session_resources.subtask_rank()}.txt"
                    errfile = f"stderr_rank{session_resources.subtask_rank()}.txt"
                with redirect_stdio(FD.STDOUT, os.path.join(workdir, outfile)):
                    with redirect_stdio(FD.STDERR, os.path.join(workdir, errfile)):
                        with self.simulation_module_context as session:
                            session.run()
                logger.debug(
                    f"workdir[{self.ensemble_rank}] = {workdir_list[self.ensemble_rank]}"
                )
                logger.debug(
                    f"parameters[{self.ensemble_rank}] = {parameters_dict_list[self.ensemble_rank]}"
                )
                logger.debug(
                    f"runtime_args[{self.ensemble_rank}] = {runtime_args_list[self.ensemble_rank]}"
                )
                # end if _current_rank_participates

            # end scoped_resources: session_resources

        # Info from other ranks might not have been available when we originally constructed
        # the list(s)
        if base_comm:
            logger.debug(
                "Waiting for simulations to complete on all ranks before publishing results."
            )
            # Extra unused ranks will not participate in the collective work,
            # but all Python interpreter processes are expected to behave consistently
            # with respect to references to results.
            # We don't know whether there are non-participating ranks, so we make this
            # collective call, even though it may be unnecessary.
            assert isinstance(base_comm, mpi4py_Comm)
            workdir_list = base_comm.bcast(workdir_list)
            parameters_dict_list = base_comm.bcast(parameters_dict_list)
            runtime_args_list = base_comm.bcast(runtime_args_list)
        if hasattr(base_comm, "barrier"):
            # This is heavy-handed. Hopefully we can replace the explicit MPI calls with a Future
            # abstraction and only wait on individual results when they are actually consumed.
            # As of gmxapi 0.3, SubscriptionPublishingRunner takes responsibility for the full
            # ensemble results on all ranks, and checks for file existence before publishing.
            base_comm.barrier()

        # Replace the local-specific workdir value with the ensemble values.
        self.workdir = list(workdir_list)
        # Set the other output attributes.
        self.parameters = list(parameters_dict_list)
        self.runtime_args = runtime_args_list


class SubscriptionSessionResources(object):
    """Input and output run-time resources for a MDRun subscription.

    A generalization of this class is the probably the main hook for customizing the resources
    provided to the operation at run time.

    .. todo:: Better factoring of SessionResources, ResourceFactory, Director.resource_factory.
    """

    def __init__(
        self, input: LegacyImplementationSubscription, output: PublishingDataProxy
    ):
        # This is instantiated by the ResourceManager and then provided to the task runner to update
        # the mdrun reference. In the current implementation, mdrun has actually already run as a
        # gmxapi 0.0.7 task, and is provided to as the input LegacyImplementationSubscription.
        assert isinstance(input, LegacyImplementationSubscription)
        assert isinstance(output, PublishingDataProxy)
        self.output = output
        member_id = self.output._client_identifier
        # Before enabling the following, be sure we understand what is happening.
        # if member_id is None:
        #     member_id = 0
        # We don't currently keep a reference to the gmxapi 0.0.7 work. We just grab its local
        # output.
        self.workdir = input.workdir[member_id]
        self.parameters = input.parameters[member_id]
        self.runtime_args = input.runtime_args[member_id]


class SubscriptionPublishingRunnerDirector(_op.AbstractRunnerDirector):
    def __init__(self):
        # We expect the gmxapi 0.0.7 MD runner to be run on all ranks.
        self.allow_duplicate = True

    def __call__(self, resources) -> typing.Callable[[], None]:
        return SubscriptionPublishingRunner(resources=resources)


class SubscriptionPublishingRunner(object):
    """Handle execution in the gmxapi.operation context as a subscription to the gmxapi.simulation.context."""

    def __init__(self, resources: SubscriptionSessionResources):
        assert isinstance(resources, SubscriptionSessionResources)
        # Note that the resources contain a reference to a simulation ensemble that has already run.
        self.resources = resources

    def __call__(self):
        """Operation implementation in the gmxapi.operation module context."""
        publisher: PublishingDataProxy = self.resources.output
        assert isinstance(publisher, PublishingDataProxy)
        publisher.directory = self.resources.workdir
        publisher.stdout = os.path.join(self.resources.workdir, "stdout.txt")
        publisher.stderr = os.path.join(self.resources.workdir, "stderr.txt")
        publisher.parameters = self.resources.parameters
        logger.debug(
            f"Session resources have runtime_args: {self.resources.runtime_args}"
        )

        # Note: the gromacs library still does not provide a way to query the outputs
        # produced through the API!
        # The '-o' and '-cpo' values are either provided by the user through *runtime_args*
        # or hard-coded in the gmxapi::Context details.
        # TODO(#3130,#3379): Make the return value a trajectory handle rather than a file path.
        # Note: There may be some ambiguity about how best to handle append vs. noappend
        # output, and how the user interface should represent the different possible behaviors.
        trajectory = self.resources.runtime_args.get("-o", None)
        if trajectory is not None:
            trajectory = Path(trajectory)
            if not trajectory.is_absolute():
                trajectory = Path(self.resources.workdir) / trajectory
        # With `-noappend`, the output filename may not be as specified by the user,
        # and we do not have a way to query the file that is actually produced!
        if trajectory is None or not trajectory.exists():
            if trajectory is None:
                stem, suffix = "traj", ".trr"
                dir = Path(self.resources.workdir)
            else:
                trajectory = Path(trajectory)
                stem, suffix = trajectory.stem, trajectory.suffix
                dir = trajectory.parent
            candidates = dir.glob(f"{stem}*{suffix}")
            trajectory = max(
                [(candidate.stat().st_mtime, candidate) for candidate in candidates],
                default=(None, None),
                key=lambda x: x[0],
            )[1]

        # TODO(#3379): Make the return value a SimulationInput handle rather than a
        #  file path.
        checkpoint = self.resources.runtime_args.get("-cpo", "state.cpt")
        if not os.path.isabs(checkpoint):
            checkpoint = os.path.join(self.resources.workdir, checkpoint)

        if trajectory:
            path = Path(trajectory)
            if path.exists():
                # Publish the output file.
                publisher.trajectory = str(path)
            else:
                logger.info(f"Output file {trajectory} does not exist.")
                try:
                    dir = path.parent
                    contents = list(str(item) for item in dir.iterdir())
                    logger.info(f"Directory {dir} contents: " ", ".join(contents))
                except FileNotFoundError:
                    # We weren't able to get a dir listing to log.
                    pass
                # Explicitly publish a null result.
                publisher.trajectory = None
        else:
            logger.info(f'Missing output. *trajectory* is "{str(trajectory)}"')
            publisher.trajectory = None

        # Developer note: since `mdrun` is used for tasks beyond MD simulation
        # (such as energy minimization), not all output files exist in all use cases.
        if checkpoint:
            path = Path(checkpoint)
            if path.exists():
                # Publish the output file.
                publisher.checkpoint = str(path)
            else:
                logger.info(f"Output file {checkpoint} does not exist.")
                try:
                    dir = path.parent
                    contents = list(str(item) for item in dir.iterdir())
                    logger.info(f"Directory {dir} contents: " ", ".join(contents))
                except FileNotFoundError:
                    # We weren't able to get a dir listing to log.
                    pass
                # Explicitly publish a null result.
                publisher.checkpoint = None
        else:
            logger.info(f'Missing output. *checkpoint* is "{str(checkpoint)}"')
            publisher.checkpoint = None


#
# Implementation
#


class ResourceManager(gmxapi.operation.ResourceManager):
    """Manage resources for the MDRun operation in the gmxapi.operation contexts.

    Extends gmxapi.operation.ResourceManager to tolerate non-standard data payloads.
    Futures managed by this resource manager may contain additional attributes.
    """

    def future(self, name: str, description: _op.ResultDescription):
        # gmxapi 0.3 introduces some subscriptions and callbacks to synchronize results across MPI ranks.
        # It will become increasingly important to optimize out unnecessary data transfers with greater
        # awareness of data locality and data flow dependencies. This is one place to evolve additional
        # hooks during or after progress on issue #3379
        _future = super().future(name=name, description=description)
        return _future

    def data(self) -> OutputDataProxy:
        return OutputDataProxy(self)

    def update_output(self):
        """Override gmxapi.operation.ResourceManager.update_output because we handle paralellism as 0.0.7."""
        # For the moment, this is copy-pasted from gmxapi.operation.ResourceManager,
        # but the only part we need to override is the ensemble handling at `for i in range(self.ensemble_width)`
        # TODO: Reimplement as the resource factory and director for the operation target context.
        if not self.done():
            self.__operation_entrance_counter += 1
            if self.__operation_entrance_counter > 1:
                raise exceptions.ProtocolError(
                    "Bug detected: resource manager tried to execute operation twice."
                )
            with self.publishing_resources() as publishing_resources:
                # TODO: rewrite with the pattern that this block is directing and then resolving an operation in the
                #  operation's library/implementation context.

                ###
                # Note: this is the resource translation from gmxapi.operation context
                # to the dispatching runner director. It uses details of the gmxapi.operation.Context
                # and of the operation.

                # Create on all ranks.
                # Unlike gmxapi.operation.ResourceManager, here we create the input resources
                # once for the entire ensemble, rather than once per ensemble member.
                # This is because the simulation actually runs as an ensemble operation
                # (in gmxapi 0.0.7 context) in order to service the input resources of this
                # version of the mdrun operation, which in actuality merely retrieves the
                # gmxapi 0.0.7 results for the current interface.
                # Abstractions that could allow reunification with the parent implementation
                # could be asyncio or concurrent processing of the ensemble members, or a `map`
                # generalization that could be implemented in serial or parallel according to the
                # ResourceManager and task requirements.
                # TODO: Dispatch/discover this resource factory from a canonical place.
                input = LegacyImplementationSubscription(self)
                # End of action of the InputResourceDirector[Context, MdRunSubscription].
                ###

                # We are giving the director a resource that contains the subscription
                # to the dispatched work.
                for member in range(self.ensemble_width):
                    with publishing_resources.publishing_context(
                        ensemble_member=member
                    ) as output:
                        error_message = "Got {} while executing {} for operation {}."
                        try:
                            resources = self._resource_factory(
                                input=input, output=output
                            )
                        except exceptions.TypeError as e:
                            message = error_message.format(
                                e, self._resource_factory, self.operation_id
                            )
                            raise exceptions.ApiError(message) from e

                        runner = self._runner_director(resources)
                        try:
                            runner()
                        except Exception as e:
                            message = error_message.format(e, runner, self.operation_id)
                            raise exceptions.ApiError(message) from e


class StandardInputDescription(_op.InputDescription):
    """Provide the MdRun input description in gmxapi.operation Contexts."""

    # TODO: Improve fingerprinting.
    # If _make_uid can't make sufficiently unique IDs, use additional "salt".
    # Without fingerprinting, we cannot consistently hash *input* across processes,
    # but we can consistently generate integers in the same sequence the first time
    # we see each distinct input.
    _number_issued: typing.ClassVar[int] = 0
    _uids: typing.ClassVar[typing.MutableMapping[int, bytes]] = {}

    @classmethod
    def _make_uid(cls, input) -> str:
        # WARNING: The built-in hash will use memory locations, and so will not be consistent across
        # process ranks, even if the input should be the same.
        # Note also that that hash() has internal salt that is regenerated for every process.
        # TODO: Use input fingerprint for more useful identification.
        # 8 bytes is probably enough.
        if input not in cls._uids:
            uid = hashlib.md5(f"mdrun{cls._number_issued}".encode("utf8")).digest()
            existing = [key for key, value in cls._uids.items() if uid == value]
            if existing:
                raise exceptions.ProtocolError(
                    f"Calculated uid {uid.hex()} for input {input}, but uid is already used by "
                    + ", ".join(repr(edge) for edge in existing)
                    + "."
                )
            cls._uids[input] = uid
            cls._number_issued += 1
        else:
            uid = cls._uids[input]
            logger.debug(f"Reissuing uid for mdrun({input}): {uid.hex()}")
        new_uid = f"mdrun_{uid.hex()}"
        return new_uid

    def signature(self) -> _op.InputCollectionDescription:
        return _input

    def make_uid(self, input: _op.DataEdge) -> str:
        assert isinstance(input, _op.DataEdge)
        return self._make_uid(input)


class RegisteredOperation(_op.OperationImplementation, metaclass=_op.OperationMeta):
    """Provide the gmxapi compatible MDRun implementation."""

    # This is a class method to allow the class object to be used in gmxapi.operation._make_registry_key
    @classmethod
    def name(cls) -> str:
        """Canonical name for the operation."""
        return "mdrun"

    @classmethod
    def namespace(cls) -> str:
        """modify_input is importable from the gmxapi module."""
        return "gmxapi"

    @classmethod
    def director(cls, context: gmxapi.abc.Context) -> gmxapi.abc.OperationDirector:
        # Currently, we only have a Directory for the gmxapi.operation.Context
        if isinstance(context, _op.Context):
            return StandardDirector(context)


class StandardOperationHandle(_op.AbstractOperation):
    """Handle used in Python UI or gmxapi.operation Contexts."""

    def __init__(self, resource_manager: ResourceManager):
        self.__resource_manager = resource_manager

    def run(self):
        self.__resource_manager.update_output()

    @property
    def output(self) -> OutputDataProxy:
        return self.__resource_manager.data()


class StandardDirector(gmxapi.abc.OperationDirector):
    """Direct the instantiation of a mdrun node in a gmxapi.operation Context.

    .. todo:: Compose this behavior in a more generic class.

    .. todo:: Describe where instances live.
    """

    def __init__(self, context: _op.Context):
        if not isinstance(context, _op.Context):
            raise gmxapi.exceptions.ValueError(
                "StandardDirector requires a gmxapi.operation Context."
            )
        self.context = context

    def __call__(
        self, resources: _op.DataSourceCollection, label: str = None
    ) -> StandardOperationHandle:
        builder = self.context.node_builder(operation=RegisteredOperation, label=label)

        builder.set_resource_factory(SubscriptionSessionResources)
        builder.set_input_description(StandardInputDescription())
        builder.set_handle(StandardOperationHandle)
        # The runner in the gmxapi.operation context is the servicer for the legacy context.
        builder.set_runner_director(SubscriptionPublishingRunnerDirector())
        builder.set_output_factory(_output_factory)
        builder.set_resource_manager(ResourceManager)

        # Note: we have not yet done any dispatching based on *resources*. We should
        # translate the resources provided into the form that the Context can subscribe to
        # using the dispatching resource_factory. In the second draft, this operation
        # is dispatched to a SimulationModuleContext, which can be subscribed directly
        # to sources that are either literal filenames or gmxapi.simulation sources,
        # while standard Futures can be resolved in the standard context.
        #
        assert isinstance(resources, _op.DataSourceCollection)
        for name, source in resources.items():
            builder.add_input(name, source)

        handle = builder.build()
        assert isinstance(handle, StandardOperationHandle)
        return handle

    def handle_type(self, context: gmxapi.abc.Context):
        return StandardOperationHandle

    # Developer note: the Director instance is a convenient place to get a dispatching
    # factory. The Director may become generic or more universal, but the resource_factory
    # would likely not be typed on the generic parameters of the Director class.
    # Instead, it is likely a generic function with its own TypeVar parameters.
    def resource_factory(
        self,
        source: typing.Union[gmxapi.abc.Context, ModuleObject, None],
        target: gmxapi.abc.Context = None,
    ):
        """Get a resource factory for use in the target context.

        The returned factory takes input from the source context and provides it in a form
        usable by the operation builder in the target context.
        """
        # Distinguish between the UIContext, in which input is in the form
        # of function call arguments, and the StandardContext, implemented in
        # gmxapi.operation. UIContext is probably a virtual context that is
        # asserted by callers in order to get a factory that normalizes UI input
        # for the StandardContext.
        #
        if target is None:
            target = self.context
        if source is None:
            # `source is None` implies source is coming from UI.
            if isinstance(target, _op.Context):
                # Return a factory that can bind to function call arguments to produce a DataSourceCollection.
                return _standard_node_resource_factory
        if isinstance(source, _op.Context):
            # The source is a gmxapi.operation.Context when the operation is being evaluated through
            # a gmxapi.operation.ResourceManager. i.e. at run time.
            return SubscriptionSessionResources
        if isinstance(source, ModuleObject):
            # In the UI context, the source may be a ModuleObject instead of `None`, per `mdrun`
            # defined below.
            if isinstance(target, _op.Context):
                # We are creating a node in gmxapi.operation.Context from another gmxapi.simulation operation.
                # This means that we want to subscribe to the subcontext instead of the gmxapi.operation.Context.
                # In the first draft, though, we just access a special payload.
                # Return a factory that will consume *_simulation_input* and *parameters*
                # members of a received object.
                logger.info("Building mdrun operation from source {}".format(source))

                def simulation_input_workaround(_simulation_input, runtime_args):
                    """Allows support for the as-yet-undefined SimulationInput resource.

                    Also supports a user interface in which the OutputDataProxy is not an
                    explicit attribute. See https://gitlab.com/gromacs/gromacs/-/issues/3174
                    """
                    source = _simulation_input
                    # Accept either an OutputDataProxy with appropriate members, or an abject
                    # that provides such an OutputDataProxy.
                    if hasattr(source, "output"):
                        source = _simulation_input.output
                    assert hasattr(source, "_simulation_input")
                    assert hasattr(source, "parameters")
                    logger.info(
                        "mdrun receiving input {}: {}".format(
                            source._simulation_input.name,
                            source._simulation_input.description,
                        )
                    )
                    source_collection = _input.bind(
                        _simulation_input=source._simulation_input,
                        parameters=source.parameters,
                        runtime_args=runtime_args,
                    )
                    logger.info(
                        "mdrun input bound as source collection {}".format(
                            source_collection
                        )
                    )
                    return source_collection

                return simulation_input_workaround

        raise gmxapi.exceptions.ValueError(
            f"No dispatching from {source} context to {target}"
        )


def mdrun(
    input,
    runtime_args: typing.Union[dict, typing.Sequence[dict]] = None,
    label: str = None,
    context=None,
):
    """MD simulation operation.

    Arguments:
        input : valid simulation input
        runtime_args (dict): command line flags and arguments to be passed to mdrun (optional)

    Returns:
        runnable operation to perform the specified simulation

    See :py:class:`~gmxapi.simulation.mdrun.OutputDataProxy` for members of the
    *output* attribute.

    *input* may be a TPR file name or an object providing the SimulationInput interface.

    *runtime_args* allows an optional dictionary of mdrun options, using the option flag
    (including the leading hyphen ``-``) as the dictionary key.
    For mdrun command line options that do not take a value (e.g. ``-noappend``),
    use ``None`` as the dictionary value.

    Warning:
        Run time argument processing does not have Python bindings at this time.
        key-value pairs are passed as plain text to the underlying library.
        Usage errors can be hard to discover. Refer to the MD log file in the output
        directory for messages regarding argument processing.

        Note, in particular, that the available ``mdrun`` arguments can depend on the
        GROMACS build configuration, such as whether an MPI library or thread-MPI
        is enabled.

    See Also:
        The :doc:`/onlinehelp/gmx-mdrun` command line tool.

    """
    # The job of this function is to arrange for the creation of a workflow node
    # by transforming arguments into a DataSourceCollection and providing the
    # collection to a Director.
    # The Director is specific to the Operation (this module) but could be generated
    # with the help of simple metaprogramming in the future (e.g. just declare
    # requirements or helpers in a class definition).
    # Ref: gmxapi.abc.NodeBuilder and gmxapi.operation.NodeBuilder

    if runtime_args is None:
        runtime_args = {}

    handle_context = context
    if handle_context is not None:
        raise gmxapi.exceptions.MissingImplementationError(
            "context must be None. This factory is only for the Python UI right now."
        )

    target_context = _op.current_context()
    assert isinstance(target_context, _op.Context)
    # Get a director that will create a node in the standard context.
    node_director: StandardDirector = _op._get_operation_director(
        RegisteredOperation, context=target_context
    )
    assert isinstance(node_director, StandardDirector)
    # TODO: refine this protocol
    assert handle_context is None
    # Examine the input.
    if isinstance(input, ModuleObject):
        # The input is from read_tpr or modify_input.
        source_context = input
    else:
        # Allow automatic dispatching
        source_context = None

    # Depending on dispatching, the ResourceFactory may be one of several functor types.
    # TODO: Use a standard helper function to pass data and resources to the NodeDirector.
    #  Module-specific input processing logic should be explicit in this user-facing function
    #  or composed into the NodeDirector for this operation.
    resource_factory = node_director.resource_factory(
        source=source_context, target=target_context
    )
    # The resource_factory will ultimately produce a DataSourceCollection with help from the
    # InputCollectionDescription defined for this module. Per InputCollectionDescription.bind()
    # and the POSITIONAL_OR_KEYWORD property, the `input` could be provided as a positional
    # argument (as it was previously) and get bound to `_simulation_input`. Despite the
    # provisional nature of the `_simulation_input` input, this code is more readable when we
    # use a named parameter.
    resources: _op.DataSourceCollection = resource_factory(
        _simulation_input=input, runtime_args=runtime_args
    )
    handle = node_director(resources=resources, label=label)
    # Note: One effect of the assertions above is to help the type checker infer
    # the return type of the handle. It is hard to convince the type checker that
    # the return value of the node builder is up-cast. We might be able to resolve
    # this by making both get_operation_director and ReadTprImplementation
    # generics of the handle type, or using the handle type as the key for
    # get_operation_director.
    return handle



#
# This file is part of the GROMACS molecular simulation package.
#
# Copyright 2019- The GROMACS Authors
# and the project initiators Erik Lindahl, Berk Hess and David van der Spoel.
# Consult the AUTHORS/COPYING files and https://www.gromacs.org for details.
#
# GROMACS is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# GROMACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with GROMACS; if not, see
# https://www.gnu.org/licenses, or write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
#
# If you want to redistribute modifications to GROMACS, please
# consider that scientific software is very special. Version
# control is crucial - bugs must be traceable. We will be happy to
# consider code for inclusion in the official distribution, but
# derived work must not be called official GROMACS. Details are found
# in the README & COPYING files - if they are missing, get the
# official version at https://www.gromacs.org.
#
# To help us fund GROMACS development, we humbly ask that you cite
# the research papers on the package. Check out https://www.gromacs.org.

"""read_tpr operation module

Provides implementation classes and user interface for gmxapi.read_tpr.
"""

import inspect
import typing

import gmxapi
import gmxapi.abc
import gmxapi.exceptions
import gmxapi.operation as _op
import gmxapi.typing
from gmxapi.operation import InputCollectionDescription
from gmxapi.simulation.abc import ModuleObject

from . import fileio

# Initialize module-level logger
from gmxapi import logger as root_logger

logger = root_logger.getChild("read_tpr")
logger.info("Importing {}".format(__name__))


#
# Interface classes and internal details
#

# TODO: The output of read_tpr and modify_input should be the same and part of the
#  simulation module specification. Such output is either a special type of output proxy
#  or Future.
_output_descriptors = (
    _op.OutputDataDescriptor("_simulation_input", str),
    _op.OutputDataDescriptor("parameters", dict),
)
_publishing_descriptors = {
    desc._name: gmxapi.operation.Publisher(desc._name, desc._dtype)
    for desc in _output_descriptors
}
_output = _op.OutputCollectionDescription(
    **{descriptor._name: descriptor._dtype for descriptor in _output_descriptors}
)


class OutputDataProxy(ModuleObject, _op.DataProxyBase, descriptors=_output_descriptors):
    """Implement the 'output' attribute of `read_tpr` operations.

    Attributes:
        parameters: Dictionary of simulation parameters.

    Additionally (through an unspecified interface), the object serves as a
    complete simulation input to other gmxapi operations.
    """

    def __init__(self, *args, **kwargs):
        _op.DataProxyBase.__init__(self, *args, **kwargs)


class PublishingDataProxy(_op.DataProxyBase, descriptors=_publishing_descriptors):
    """Manage output resource updates for ReadTpr operation."""


_output_factory = _op.OutputFactory(
    output_proxy=OutputDataProxy,
    output_description=_output,
    publishing_data_proxy=PublishingDataProxy,
)


class SessionResources(object):
    """Input and output run-time resources for a ReadTpr operation."""

    def __init__(self, tpr_filename, publisher: PublishingDataProxy):
        self.tpr_object = fileio.TprFile(filename=tpr_filename, mode="r")
        self.output = publisher


#
# Helpers
#


_input = _op.InputCollectionDescription(
    [
        (
            "filename",
            inspect.Parameter(
                "filename", inspect.Parameter.POSITIONAL_OR_KEYWORD, annotation=str
            ),
        )
    ]
)


# TODO: Clarify. The actual input and output arguments passed are customized for this operation.
def _session_resource_factory(
    input: _op.InputPack, output: "PublishingDataProxy", **kwargs
) -> SessionResources:
    """Translate resources from the gmxapi.operation Context to the ReadTpr implementation."""
    # TODO: Either get rid of **kwargs or clarify the roadmap and timeline for doing so.
    filename = input.kwargs["filename"]
    return SessionResources(tpr_filename=filename, publisher=output)


def _standard_node_resource_factory(*args, **kwargs):
    """Translate Python UI input to the gmxapi.operation node builder inputs."""
    return _input.bind(*args, **kwargs)


def _run(resources: SessionResources):
    """Operation implementation in the gmxapi.operation module context."""
    # TODO: Implement for other source/target contexts. We don't always need to
    #  produce all outputs.
    with resources.tpr_object as fh:
        params = fh._tprFileHandle.params().extract()
        resources.output.parameters = params
        resources.output._simulation_input = fh.filename


#
# Implementation
#


# Note: we borrow the implementation from operation.ResourceManager for now,
# but in the future we want the implementations to either be decoupled or
# for implementations in a given context to be coupled to details that are clearly
# and explicitly related to that context. Right now, operation.ResourceManager
# is tied to the implementation of Contexts in gmxapi.operation, but that is not
# sufficiently clear and explicit.
class ResourceManager(gmxapi.operation.ResourceManager):
    """Manage resources for the ReadTpr operation in the gmxapi.operation contexts.

    Extends gmxapi.operation.ResourceManager to tolerate non-standard data payloads.
    Futures managed by this resource manager may contain additional attributes.
    """

    def future(self, name: str, description: _op.ResultDescription):
        tpr_future = super().future(name=name, description=description)
        return tpr_future

    def data(self) -> OutputDataProxy:
        return OutputDataProxy(self)


# TODO: Consider making Generic in source and target context type variables,
#  or leave unspecified and use generic function or pair of single_dispatch functions.
#  Need to know the right home for input_description, if needed.
class ResourceFactory(gmxapi.abc.ResourceFactory):
    """ReadTpr resource factory.

    Generic class for creating resources passed to read_tpr implementation details.
    Dispatching may occur based on the source and target Context of factory action.
    """

    def __init__(self, target_context, source_context):
        """Initialize an instance to support read_tpr action.

        Arguments:
            source_context: The source of the resources in the calling scope.
            target_context: The Context in which the factory product will be consumed.

        """
        self.source_context = source_context  # Determine input form of *create* method.
        self.target_context = target_context  # Context in which resources are consumed.

    # TODO: clean up the dispatching. What to do when input comes from multiple sources?
    # Use a typing overload or a single-dispatch functor for clearer input/result typing.
    def __call__(self, *args, **kwargs):
        """Create the resource product.

        Dispatch to appropriate factory functionality based on the *source_context*
        and *target_context*.
        """
        # context is used for dispatching and annotates the Context of the other arguments.
        # context == None implies inputs are from Python UI.
        if self.source_context is None:
            if isinstance(self.target_context, _op.Context):
                return _standard_node_resource_factory(*args, **kwargs)
        if isinstance(self.source_context, _op.Context):
            # TODO: Check whether the consumer is a Context.NodeBuilder or an operation runner.
            # We don't yet use this dispatcher for building nodes, so assume we are launching a session.
            assert "input" in kwargs
            assert "output" in kwargs
            return _session_resource_factory(
                input=kwargs["input"], output=kwargs["output"]
            )
        raise gmxapi.exceptions.MissingImplementationError(
            "No translation from {} context to {}".format(
                self.source_context, self.target_context
            )
        )

    @typing.overload
    def input_description(self, context: _op.Context) -> _op.InputDescription: ...

    def input_description(self, context: gmxapi.abc.Context):
        """Get an input description usable in the indicated Context.

        Arguments:
            context: Context for which to dispatch the generation of input description.

        Overrides gmxapi.abc.ResourceFactory for collaboration between this
        resource factory and a target Context.
        """
        # TODO: Clarify exposure and scope as gmxapi.abc.ResourceFactory is refined.
        # The expected use case is that a Context implementation may consult the
        # input_description when preparing input to a resource factory, or even
        # when determining in which Context a resource should be created or an
        # operation dispatched. It seems sensible that the ResourceFactory should
        # be able to describe its possible inputs. Instance functions are the
        # most common and simple Python detail, so it makes sense for this to be
        # an instance method instead of classmethod or staticmethod, until shown
        # otherwise.
        # Also note that ResourceFactory is likely to be a Generic class or a
        # container for composed functionality. Customizing method signatures
        # based on instance data is more brittle than behavior determined at the
        # class level. The behavior of this function is determined at the class
        # level to be a dispatcher, which may utilize instance data, but as an
        # implementation detail that is not the business of the caller.
        # This should probably be implemented in terms of the standard Python
        # functools.single_dispatch generic function, but it is cleaner if the
        # method itself is not generic beyond the level of typing overloads.
        if isinstance(context, _op.Context):
            return StandardInputDescription()
        raise gmxapi.exceptions.MissingImplementationError(
            "No input description available for {} context".format(context)
        )


class StandardInputDescription(_op.InputDescription):
    """Provide the ReadTpr input description in gmxapi.operation Contexts."""

    # TODO: Improve fingerprinting.
    # If _make_uid can't make sufficiently unique IDs, use additional "salt".
    # Without fingerprinting, we cannot consistently hash *input* across processes,
    # but we can consistently generate integers in the same sequence the first time
    # we see each distinct input.
    _next_uid: typing.ClassVar[int] = 0
    _uids: typing.ClassVar[typing.MutableMapping[int, int]] = {}

    @classmethod
    def _make_uid(cls, input) -> str:
        # TODO: Use input fingerprint for more useful identification.
        # WARNING: The built-in hash will use memory locations, and so will not be consistent across
        # process ranks, even if the input should be the same.
        salt = hash(input)
        if salt not in cls._uids:
            cls._uids[salt] = cls._next_uid
            cls._next_uid += 1
        else:
            logger.debug(f"Reissuing uid for read_tpr({input}): {cls._uids[salt]}")
        new_uid = "read_tpr_{}".format(cls._uids[salt])
        return new_uid

    def signature(self) -> InputCollectionDescription:
        return _input

    def make_uid(self, input: _op.DataEdge) -> str:
        assert isinstance(input, _op.DataEdge)
        return self._make_uid(input)


class RegisteredOperation(_op.OperationImplementation, metaclass=_op.OperationMeta):
    """Provide the gmxapi compatible ReadTpr implementation."""

    # This is a class method to allow the class object to be used in gmxapi.operation._make_registry_key
    @classmethod
    def name(self) -> str:
        """Canonical name for the operation."""
        return "read_tpr"

    @classmethod
    def namespace(self) -> str:
        """read_tpr is importable from the gmxapi module."""
        return "gmxapi"

    @classmethod
    def director(cls, context: gmxapi.abc.Context) -> gmxapi.abc.OperationDirector:
        if isinstance(context, _op.Context):
            return StandardDirector(context)
        raise gmxapi.exceptions.MissingImplementationError(
            "No dispatcher for context {} of type {}".format(context, type(context))
        )


class StandardOperationHandle(_op.AbstractOperation, ModuleObject):
    """Handle used in Python UI or gmxapi.operation Contexts."""

    def __init__(self, resource_manager: ResourceManager):
        self.__resource_manager = resource_manager

    def run(self):
        self.__resource_manager.update_output()

    @property
    def output(self) -> OutputDataProxy:
        return self.__resource_manager.data()


class StandardDirector(gmxapi.abc.OperationDirector):
    """Direct the instantiation of a read_tpr node in a gmxapi.operation Context.

    .. todo:: Compose this behavior in a more generic class.

    .. todo:: Describe where instances live.
    """

    def __init__(self, context: _op.Context):
        if not isinstance(context, _op.Context):
            raise gmxapi.exceptions.ValueError(
                "StandardDirector requires a gmxapi.operation Context."
            )
        self.context = context

    def __call__(
        self, resources: _op.DataSourceCollection, label: str
    ) -> StandardOperationHandle:
        builder = self.context.node_builder(operation=RegisteredOperation, label=label)

        builder.set_resource_factory(_session_resource_factory)
        builder.set_input_description(StandardInputDescription())
        builder.set_handle(StandardOperationHandle)

        runner_director = _op.RunnerDirector(runner=_run, allow_duplicate=True)

        builder.set_runner_director(runner_director)
        builder.set_output_factory(_output_factory)

        # Note: we have not yet done any dispatching based on *resources*. We should
        # translate the resources provided into the form that the Context can subscribe to
        # using the dispatching resource_factory. In the second draft, this operation
        # is dispatched to a SimulationModuleContext, which can be subscribed directly
        # to sources that are either literal filenames or gmxapi.simulation sources,
        # while standard Futures can be resolved in the standard context.
        #
        assert isinstance(resources, _op.DataSourceCollection)
        assert "filename" in resources
        builder.add_input("filename", resources["filename"])

        handle = builder.build()
        assert isinstance(handle, StandardOperationHandle)
        return handle

    def handle_type(self, context: gmxapi.abc.Context):
        return StandardOperationHandle

    def resource_factory(
        self,
        source: typing.Union[gmxapi.abc.Context, None],
        target: gmxapi.abc.Context = None,
    ) -> ResourceFactory:
        # Distinguish between the UIContext, in which input is in the form
        # of function call arguments, and the StandardContext, implemented in
        # gmxapi.operation. UIContext is probably a virtual context that is
        # asserted by callers in order to get a factory that normalizes UI input
        # for the StandardContext.
        #
        if target is None:
            target = self.context
        if source is None:
            if isinstance(target, _op.Context):
                return ResourceFactory(target_context=target, source_context=source)
        if isinstance(source, _op.Context):
            return ResourceFactory(target_context=target, source_context=source)
        raise gmxapi.exceptions.ValueError(
            "No dispatching from {} context to {}".format(source, target)
        )


def read_tpr(filename, label: str = None, context=None):
    """Get simulation input from a TPR file.

    Arguments:
        filename: input file name
        label: optional human-readable label with which to tag the new node
        context: Context in which to return a handle to the new node.
                 Use default (None) for Python scripting interface

    Returns:
        Reference (handle) to the new operation instance (node).

    See :py:class:`~gmxapi.simulation.read_tpr.OutputDataProxy` for members of
    the *output* attribute.
    """
    handle_context = context
    if handle_context is not None:
        raise gmxapi.exceptions.MissingImplementationError(
            "context must be None. This factory is only for the Python UI right now."
        )

    # 1. Handle node creation in the scripting interface.
    # When *context* input is None, dispatch to the current Context. Confirm that
    # it is a standard context from the gmxapi.operation module, translate the
    # input into that context, and create the node.

    # 2. Dispatch to SimulationModuleContext.
    # Operation is not fully implemented in gmxapi.operation context. When creating
    # a node in a gmxapi.operation context, dispatch and subscribe to a SimulationModuleContext,
    # in which

    # 3. Handle consumption in SimulationModuleContext.
    # When a consuming operation is native to the SimulationModuleContext,
    # detect that a richer interface is available and use it. Possible implementation:
    # Chain of Responsibility: subscribe() is serviced by the Context "closest"
    # to the source. subscriber is the most native compatible Context for the consuming operation
    # but decays to the context in which the handle is being created. subscribe()
    # can accept a callback function and an indication of the form of the message
    # to send (a consuming Context or ResourceFactory).

    # TODO: Other types of input, such as File placeholders.

    target_context = _op.current_context()
    assert isinstance(target_context, _op.Context)
    # Get a director that will create a node in the standard context.
    node_director = _op._get_operation_director(
        RegisteredOperation, context=target_context
    )
    assert isinstance(node_director, StandardDirector)
    # TODO: refine this protocol
    assert handle_context is None
    resource_factory = node_director.resource_factory(
        source=handle_context, target=target_context
    )
    resources = resource_factory(filename=filename)
    handle = node_director(resources=resources, label=label)
    # Note: One effect of the assertions above is to help the type checker infer
    # the return type of the handle. It is hard to convince the type checker that
    # the return value of the node builder is up-cast. We might be able to resolve
    # this by making both get_operation_director and ReadTprImplementation
    # generics of the handle type, or using the handle type as the key for
    # get_operation_director.
    return handle



#
# This file is part of the GROMACS molecular simulation package.
#
# Copyright 2019- The GROMACS Authors
# and the project initiators Erik Lindahl, Berk Hess and David van der Spoel.
# Consult the AUTHORS/COPYING files and https://www.gromacs.org for details.
#
# GROMACS is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# GROMACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with GROMACS; if not, see
# https://www.gnu.org/licenses, or write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
#
# If you want to redistribute modifications to GROMACS, please
# consider that scientific software is very special. Version
# control is crucial - bugs must be traceable. We will be happy to
# consider code for inclusion in the official distribution, but
# derived work must not be called official GROMACS. Details are found
# in the README & COPYING files - if they are missing, get the
# official version at https://www.gromacs.org.
#
# To help us fund GROMACS development, we humbly ask that you cite
# the research papers on the package. Check out https://www.gromacs.org.

"""Provide the high-level interface to the file i/o behaviors in the package.
"""
# The submodule is named "fileio" instead of "io" to avoid a
# namespace collision with a standard Python module on the default path.

import typing

__all__ = ["TprFile", "read_tpr", "write_tpr_file"]

import os

from gmxapi import exceptions


class TprFile(object):
    """Handle to a GROMACS simulation run input file.

    TprFile objects do not have a public interface. The class is used internally
    to manage simulation input data structures.

    Attributes:
        filename (str): Name of the file with which the object was initialized.
        mode: File access mode from object creation.

    """

    def __init__(self, filename: str = None, mode: str = "r"):
        """Open a TPR file.

        File access mode is indicated by 'r' for read-only access.

        Args:
            filename (str): Path to a run input file (e.g. 'myfile.tpr')
            mode (str): File access mode.

        Note:
            Currently, TPR files are read-only from the Python interface.

        """
        if filename is None:
            raise exceptions.UsageError(
                "TprFile objects must be associated with a file."
            )
        if mode != "r":
            raise exceptions.UsageError("TPR files only support read-only access.")
        self.mode = mode
        self.filename = filename
        self._tprFileHandle = None

    def close(self):
        # self._tprFileHandle.close()
        self._tprFileHandle = None

    def __repr__(self):
        return "{}('{}', '{}')".format(
            self.__class__.__name__, self.filename, self.mode
        )

    def __enter__(self):
        import gmxapi._gmxapi as _gmxapi

        self._tprFileHandle = _gmxapi.read_tprfile(self.filename)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return


class _NodeOutput(object):
    """Implement the `output` attribute of a simulation input node.

    Attributes:
        parameters: Simulation parameters for (re)written TPR file.
        structure: Atomic details (not yet implemented)
        topology: Molecular force field details (not yet implemented)
        state: Simulation state information (not yet implemented)

    """

    def __init__(self, parameters=None, structure=None, topology=None, state=None):
        """Initialize getters for output ports."""
        self.__tprfile = parameters

    @property
    def parameters(self):
        with self.__tprfile as fh:
            params = fh._tprFileHandle.params()
        return params

    @property
    def structure(self):
        raise exceptions.ApiError("property not implemented.")

    @property
    def topology(self):
        raise exceptions.ApiError("property not implemented.")

    @property
    def state(self):
        raise exceptions.ApiError("property not implemented.")


class _SimulationInput(object):
    """
    Simulation input interface for a TPR file read by gmx.fileio.read_tpr()

    Attributes:
        parameters: Simulation parameters for (re)written TPR file.
        structure: Atomic details (not yet implemented)
        topology: Molecular force field details (not yet implemented)
        state: Simulation state information (not yet implemented)

    """

    def __init__(self, tprfile: typing.Union[str, TprFile]):
        if not isinstance(tprfile, TprFile):
            try:
                tprfile = TprFile(tprfile)
            except Exception as e:
                # This class is an implementation detail of TPR file I/O...
                raise exceptions.ApiError("Must be initialized from a TprFile.") from e
        assert isinstance(tprfile, TprFile)
        self.__tprfile = tprfile
        self.__parameters = None

    @property
    def parameters(self):
        if self.__parameters is None:
            with self.__tprfile as fh:
                self.__parameters = fh._tprFileHandle.params()
        return self.__parameters

    @property
    def structure(self):
        raise exceptions.ApiError("property not implemented.")

    @property
    def topology(self):
        raise exceptions.ApiError("property not implemented.")

    @property
    def state(self):
        raise exceptions.ApiError("property not implemented.")


def read_tpr(tprfile: typing.Union[str, TprFile]):
    """
    Get a simulation input object from a TPR run input file.

    Arguments:
        tprfile : TPR input object or filename

    Returns:
         simulation input object

    The returned object may be inspected by the user. Simulation input parameters
    may be extracted through the `parameters` attribute.

    Example:
        >>> sim_input = gmx.fileio.read_tpr(tprfile=tprfilename)
        >>> params = sim_input.parameters.extract()
        >>> print(params['init-step'])
        0

    Supports the `read_tpr` gmxapi work graph operation. (not yet implemented)
    """
    if not isinstance(tprfile, TprFile):
        try:
            tprfile = TprFile(os.fsencode(tprfile), mode="r")
        except Exception as e:
            raise exceptions.UsageError("TPR object or file name is required.") from e

    return _SimulationInput(tprfile)


# In initial implementation, we extract the entire TPR file contents through the
# TPR-backed GmxMdParams implementation.
# Note: this function is not consistent with a gmxapi operation.
def write_tpr_file(output, input=None):
    """
    Create a new TPR file, combining user-provided input.

    .. versionadded:: 0.0.8
        Initial version of this tool does not know how to generate a valid simulation
        run input file from scratch, so it requires input derived from an already valid
        TPR file.

    The simulation input object should provide the gmx simulation_input interface,
    with output ports for `parameters`, `structure`, `topology`, and `state`, such
    as a TprFileHandle

    Arguments:
        output : TPR file name to write.
        input : simulation input data from which to write a simulation run input file.

    Use this function to write a new TPR file with data updated from an
    existing TPR file. Keyword arguments are objects that can provide gmxapi
    compatible access to the necessary simulation input data.

    In the initial version, data must originate from an existing TPR file, and
    only simulation parameters may be rewritten. See gmx.fileio.read_tpr()

    Example:
        >>> sim_input = gmx.fileio.read_tpr(tprfile=tprfilename)
        >>> sim_input.parameters.set('init-step', 1)
        >>> gmx.fileio.write_tpr_file(newfilename, input=sim_input)

    Warning:
        The interface is in flux.

    TODO:
        Be consistent about terminology for "simulation state".
        We are currently using "simulation state" to refer both to the aggregate of
        data (superset) necessary to launch or continue a simulation _and_ to the
        extra data (subset) necessary to capture full program state, beyond the
        model/method input parameters and current phase space coordinates. Maybe we
        shouldn't expose that as a separate user-accessible object and should instead
        make it an implementation detail of a wrapper object that has standard
        interfaces for the non-implementation-dependent encapsulated data.

    Returns:
        TBD : possibly a completion condition of some sort and/or handle to the new File
    """
    import gmxapi._gmxapi as _gmxapi

    # TODO: (Data model) Decide how to find output data sources.
    if not hasattr(input, "parameters"):
        if hasattr(input, "output"):
            if hasattr(input.output, "parameters"):
                parameters = input.output.parameters
            else:
                raise ValueError("Need output.parameters")
        else:
            raise ValueError("Need output.parameters")
    else:
        parameters = input.parameters

    if not isinstance(parameters, _gmxapi.SimulationParameters):
        raise exceptions.TypeError(
            "You must provide a gmx.core.SimulationParameters object to `parameters` as input."
        )
    _gmxapi.write_tprfile(output, parameters)



#
# This file is part of the GROMACS molecular simulation package.
#
# Copyright 2019- The GROMACS Authors
# and the project initiators Erik Lindahl, Berk Hess and David van der Spoel.
# Consult the AUTHORS/COPYING files and https://www.gromacs.org for details.
#
# GROMACS is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# GROMACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with GROMACS; if not, see
# https://www.gnu.org/licenses, or write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
#
# If you want to redistribute modifications to GROMACS, please
# consider that scientific software is very special. Version
# control is crucial - bugs must be traceable. We will be happy to
# consider code for inclusion in the official distribution, but
# derived work must not be called official GROMACS. Details are found
# in the README & COPYING files - if they are missing, get the
# official version at https://www.gromacs.org.
#
# To help us fund GROMACS development, we humbly ask that you cite
# the research papers on the package. Check out https://www.gromacs.org.

"""Reusable definitions for test modules.

Provides utilities and pytest fixtures for gmxapi and GROMACS tests.

To load these facilities in a pytest environment, set a `pytest_plugins`
variable in a conftest.py
(Reference https://docs.pytest.org/en/latest/writing_plugins.html#requiring-loading-plugins-in-a-test-module-or-conftest-file)

    pytest_plugins = "gmxapi.testsupport"

.. seealso:: https://docs.pytest.org/en/latest/plugins.html#findpluginname

.. todo:: Consider moving this to a separate optional package.
"""

import os
import warnings
from contextlib import contextmanager

import pytest

mpi_status = "Test requires mpi4py managing 2 MPI ranks."
skip_mpi = False
rank_number = 0
comm_size = 1
rank_tag = ""
comm = None
try:
    from mpi4py import MPI

    if not MPI.Is_initialized():
        skip_mpi = True
        mpi_status += " MPI is not initialized"
    else:
        comm = MPI.COMM_WORLD
        if comm.Get_size() < 2:
            skip_mpi = True
            mpi_status += " MPI context is too small."
        else:
            rank_number = comm.Get_rank()
            comm_size = comm.Get_size()
except ImportError:
    MPI = None
    skip_mpi = True
    mpi_status += " mpi4py is not available."


def pytest_configure(config):
    config.addinivalue_line(
        "markers", "withmpi_only: test requires mpi4py managing 2 MPI ranks."
    )


def pytest_runtest_setup(item):
    # Handle the withmpi_only marker.
    for _ in item.iter_markers(name="withmpi_only"):
        if skip_mpi:
            pytest.skip(mpi_status)
        # The API uses iteration because markers may be duplicated, but we only
        # care about whether 'withmpi_only' occurs at all.
        break


def pytest_addoption(parser):
    """Add command-line user options for the pytest invocation."""
    # TODO(#4345): Remove `--rm` option after 2023 release.
    parser.addoption(
        "--rm",
        action="store",
        default="always",
        choices=["always", "never", "success"],
        help="No longer used. See https://docs.pytest.org/en/latest/how-to/tmp_path.html",
    )
    parser.addoption(
        "--threads",
        type=int,
        help="Maximum number of threads per process per gmxapi session.",
    )
    parser.addoption(
        "--pydevd",
        type=str,
        action="store",
        nargs="?",
        const="pydevd_pycharm",
        help="Attempt to connect to PyDev.Debugger using the indicated module.",
    )
    parser.addoption(
        "--pydevd-host", type=str, default="localhost", help="Set the pydevd host."
    )
    parser.addoption(
        "--pydevd-port", type=int, default=12345, help="Set the pydevd port."
    )
    parser.addoption(
        "--pydevd-rank",
        type=int,
        default=0,
        help="In MPI, specify the rank to attach to the debug server.",
    )


@pytest.fixture(scope="session", autouse=True)
def pydev_debug(request):
    """If requested, try to connect to a PyDev.Debugger backend at
    host.docker.internal:12345.

    Note: the IDE run configuration must be started before launching pytest.
    """
    pydevd_module = request.config.getoption("--pydevd")
    if pydevd_module:
        host = request.config.getoption("--pydevd-host")
        port = request.config.getoption("--pydevd-port")
        participating_rank = request.config.getoption("--pydevd-rank")
        if rank_number == participating_rank:
            try:
                import importlib

                pydevd = importlib.import_module(pydevd_module)
                settrace = getattr(pydevd, "settrace", None)
                if settrace is None:
                    raise RuntimeError(
                        f"PyDevD interface not supported. {pydevd} does not have `settrace`."
                    )
                return settrace(
                    host, port=port, stdoutToServer=True, stderrToServer=True
                )
            except ImportError:
                warnings.warn(f"{pydevd_module} not found. Ignoring `--pydevd` option.")


@pytest.fixture(scope="session")
def gmxconfig():
    from .commandline import _config

    config = _config()
    yield config


@pytest.fixture(scope="function")
def mdrun_kwargs(request, gmxconfig):
    """pytest fixture to provide a mdrun_kwargs dictionary for the mdrun ResourceManager.

    .. versionchanged:: 0.4.1
        Use function scope to avoid unexpectedly mutating the dictionary between tests.

    """
    if gmxconfig is None:
        raise RuntimeError("--threads argument requires a usable gmxconfig.json")
    arg = request.config.getoption("--threads")
    if arg is None:
        return {}
    mpi_type = gmxconfig["gmx_mpi_type"]
    if mpi_type is not None and mpi_type == "tmpi":
        kwargs = {"threads": int(arg)}
    else:
        kwargs = {}
    # TODO: (#3718) Normalize the handling of run-time arguments.
    return kwargs


@contextmanager
def scoped_chdir(dir):
    oldpath = os.getcwd()
    os.chdir(dir)
    try:
        yield dir
        # If the `with` block using scoped_chdir produces an exception, it will
        # be raised at this point in this function. We want the exception to
        # propagate out of the `with` block, but first we want to restore the
        # original working directory, so we skip `except` but provide a `finally`.
    finally:
        os.chdir(oldpath)


@pytest.fixture
def cleandir(tmp_path):
    """Provide a clean temporary working directory for a test.

    Temporary directory is created and managed as described at
    https://docs.pytest.org/en/latest/how-to/tmp_path.html

    TODO:
        Avoid or suppress ``PytestWarning: (rm_rf) error removing ...``
        from duplicated temporary directory management on multiple MPI ranks.
    """
    with scoped_chdir(tmp_path):
        yield tmp_path


@pytest.fixture(scope="session")
def gmxcli():
    from .commandline import cli_executable

    command = cli_executable()

    if command is None:
        message = (
            "Tests need 'gmx' command line tool, but could not find it on the path."
        )
        raise RuntimeError(message)
    try:
        assert os.access(command, os.X_OK)
    except Exception as E:
        raise RuntimeError(
            '"{}" is not an executable gmx wrapper program'.format(command)
        ) from E
    yield command



###########################################################################
#                                                                         #
#    physical_validation,                                                 #
#    a python package to test the physical validity of MD results         #
#                                                                         #
#    Written by Michael R. Shirts <michael.shirts@colorado.edu>           #
#               Pascal T. Merz <pascal.merz@colorado.edu>                 #
#                                                                         #
#    Copyright (C) 2012 University of Virginia                            #
#              (C) 2017 University of Colorado Boulder                    #
#                                                                         #
#    This library is free software; you can redistribute it and/or        #
#    modify it under the terms of the GNU Lesser General Public           #
#    License as published by the Free Software Foundation; either         #
#    version 2.1 of the License, or (at your option) any later version.   #
#                                                                         #
#    This library is distributed in the hope that it will be useful,      #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of       #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU    #
#    Lesser General Public License for more details.                      #
#                                                                         #
#    You should have received a copy of the GNU Lesser General Public     #
#    License along with this library; if not, write to the                #
#    Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,     #
#    Boston, MA 02110-1301 USA                                            #
#                                                                         #
###########################################################################
r"""
gromacs_parser.py
"""
import warnings
import numpy as np

from . import parser
# py2.7 compatibility
from .simulation_data import SimulationData
from .unit_data import UnitData
from .ensemble_data import EnsembleData
from .system_data import SystemData
from .observable_data import ObservableData
from .trajectory_data import TrajectoryData
# replace lines above by this when py2.7 support is dropped:
# from . import SimulationData, UnitData, EnsembleData, SystemData, ObservableData, TrajectoryData
from ..util.gromacs_interface import GromacsInterface
from ..util import error as pv_error


class GromacsParser(parser.Parser):
    """
    GromacsParser
    """

    @staticmethod
    def units():
        # Gromacs uses kJ/mol
        return UnitData(
            kb=8.314462435405199e-3,
            energy_str='kJ/mol',
            energy_conversion=1.0,
            length_str='nm',
            length_conversion=1.0,
            volume_str='nm^3',
            volume_conversion=1.0,
            temperature_str='K',
            temperature_conversion=1.0,
            pressure_str='bar',
            pressure_conversion=1.0,
            time_str='ps',
            time_conversion=1.0)

    def __init__(self, exe=None, includepath=None):
        r"""
        Create a GromacsParser object

        Parameters
        ----------
        exe: str, optional
            Path to a gmx executable (or simply the executable name, if it is in the path)
            Default: Looks for `gmx`, then for `gmx_d` in the path. If neither is found, `exe` is
                     set to None, and any parsing including simulation trajectories (`edr`, `trr`
                     and `gro` arguments in `get_simulation_data()`) will fail.
        includepath: str or List[str], optional
            Path or list of paths to location(s) of topology file. Is used for the lookup of
            `#include` statements in topologies.
            Default: None - no additional topology location. Lookup will be restricted to current
                     directory and location of the `top` file given to `get_simulation_data()`,
                     plus any include locations added to the `mdp` file.
        """
        super(GromacsParser, self).__init__()
        self.__interface = GromacsInterface(exe=exe, includepath=includepath)
        # gmx energy codes
        self.__gmx_energy_names = {'kinetic_energy': 'Kinetic-En.',
                                   'potential_energy': 'Potential',
                                   'total_energy': 'Total-Energy',
                                   'volume': 'Volume',
                                   'pressure': 'Pressure',
                                   'temperature': 'Temperature',
                                   'constant_of_motion': 'Conserved-En.'}

    def get_simulation_data(self,
                            mdp=None, top=None, edr=None,
                            trr=None, gro=None):
        r"""

        Parameters
        ----------
        mdp: str, optional
            A string pointing to a .mdp file
        top: str, optional
            A string pointing to a .top file
        edr: str, optional
            A string pointing to a .edr file
        trr: str, optional
            A string pointing to a .trr file
        gro: str, optional
            A string pointing to a .gro file (Note: if also trr is given, gro is ignored)

        Returns
        -------
        result: SimulationData
            A SimulationData filled with the results of the simulation as described by
            the provided GROMACS files.

        """
        result = SimulationData()
        result.units = self.units()

        # trajectories (might be used later for the box...)
        trajectory_dict = None
        if trr is not None:
            if gro is not None:
                warnings.warn('`trr` and `gro` given. Ignoring `gro`.')

            trajectory_dict = self.__interface.read_trr(trr)
            result.trajectory = TrajectoryData(
                trajectory_dict['position'],
                trajectory_dict['velocity'])
        elif gro is not None:
            trajectory_dict = self.__interface.read_gro(gro)
            result.trajectory = TrajectoryData(
                trajectory_dict['position'],
                trajectory_dict['velocity'])

        # simulation parameters & system
        if mdp is not None and top is not None:
            mdp_options = self.__interface.read_mdp(mdp)
            define = None
            include = None
            if 'define' in mdp_options:
                define = mdp_options['define']
            if 'include' in mdp_options:
                include = mdp_options['include']
            molecules = self.__interface.read_system_from_top(top, define=define, include=include)

            if 'dt' in mdp_options:
                result.dt = float(mdp_options['dt'])

            natoms = 0
            mass = []
            constraints_per_molec = []
            angles = ('constraints' in mdp_options and
                      mdp_options['constraints'] == 'all-angles')
            angles_h = (angles or
                        'constraints' in mdp_options and
                        mdp_options['constraints'] == 'h-angles')
            bonds = (angles_h or
                     'constraints' in mdp_options and
                     mdp_options['constraints'] == 'all-bonds')
            bonds_h = (bonds or
                       'constraints' in mdp_options and
                       mdp_options['constraints'] == 'h-bonds')

            molecule_idx = []
            next_molec = 0
            molec_bonds = []
            molec_bonds_constrained = []
            for molecule in molecules:
                natoms += molecule['nmolecs'] * molecule['natoms']
                for n in range(0, molecule['nmolecs']):
                    molecule_idx.append(next_molec)
                    next_molec += molecule['natoms']
                mass.extend(molecule['mass'] * molecule['nmolecs'])
                constraints = 0
                constrained_bonds = []
                all_bonds = molecule['bonds'] + molecule['bondsh']
                if molecule['settles']:
                    constraints = 3
                    constrained_bonds = all_bonds
                else:
                    if bonds:
                        constraints += molecule['nbonds'][0]
                        constrained_bonds.extend(molecule['bonds'])
                    if bonds_h:
                        constraints += molecule['nbonds'][1]
                        constrained_bonds.extend(molecule['bondsh'])
                    if angles:
                        constraints += molecule['nangles'][0]
                    if angles_h:
                        constraints += molecule['nangles'][1]
                constraints_per_molec.extend([constraints] * molecule['nmolecs'])
                molec_bonds.extend([all_bonds] * molecule['nmolecs'])
                molec_bonds_constrained.extend([constrained_bonds] * molecule['nmolecs'])

            system = SystemData()
            system.natoms = natoms
            system.mass = mass
            system.molecule_idx = molecule_idx
            system.nconstraints = np.sum(constraints_per_molec)
            system.nconstraints_per_molecule = constraints_per_molec
            system.ndof_reduction_tra = 3
            system.ndof_reduction_rot = 0
            if 'comm-mode' in mdp_options:
                if mdp_options['comm-mode'] == 'linear':
                    system.ndof_reduction_tra = 3
                elif mdp_options['comm-mode'] == 'angular':
                    system.ndof_reduction_tra = 3
                    system.ndof_reduction_rot = 3
                if mdp_options['comm-mode'] == 'none':
                    system.ndof_reduction_tra = 0
            system.bonds = molec_bonds
            system.constrained_bonds = molec_bonds_constrained
            result.system = system

            thermostat = ('tcoupl' in mdp_options and
                          mdp_options['tcoupl'] and
                          mdp_options['tcoupl'] != 'no')
            stochastic_dyn = ('integrator' in mdp_options and
                              mdp_options['integrator'] in ['sd', 'sd2', 'bd'])
            constant_temp = thermostat or stochastic_dyn
            temperature = None
            if constant_temp:
                ref_t = [float(t) for t in mdp_options['ref-t'].split()]
                if len(ref_t) == 1 or np.allclose(ref_t, [ref_t[0]]*len(ref_t)):
                    temperature = ref_t[0]
                else:
                    raise pv_error.InputError('mdp',
                                              'Ensemble definition ambiguous: Different t-ref values found.')

            constant_press = ('pcoupl' in mdp_options and
                              mdp_options['pcoupl'] and
                              mdp_options['pcoupl'] != 'no')
            volume = None
            pressure = None
            if constant_press:
                ref_p = [float(p) for p in mdp_options['ref-p'].split()]
                if len(ref_p) == 1 or np.allclose(ref_p, [ref_p[0]]*len(ref_p)):
                    pressure = ref_p[0]
                else:
                    raise pv_error.InputError('mdp',
                                              'Ensemble definition ambiguous: Different p-ref values found.')
            else:
                if trajectory_dict is not None:
                    box = trajectory_dict['box'][0]
                    # Different box shapes?
                    volume = box[0]*box[1]*box[2]
                else:
                    warnings.warn('Constant volume simulation with undefined volume.')

            if constant_temp and constant_press:
                ens = 'NPT'
            elif constant_temp:
                ens = 'NVT'
            else:
                ens = 'NVE'

            if ens == 'NVE':
                self.__gmx_energy_names['constant_of_motion'] = 'Total-Energy'
            else:
                self.__gmx_energy_names['constant_of_motion'] = 'Conserved-En.'

            result.ensemble = EnsembleData(
                ens,
                natoms=natoms,
                volume=volume, pressure=pressure,
                temperature=temperature
            )

        if edr is not None:
            observable_dict = self.__interface.get_quantities(edr,
                                                              self.__gmx_energy_names.values(),
                                                              args=['-dp'])

            # constant volume simulations don't write out the volume in .edr file
            if (observable_dict['Volume'] is None and
               result.ensemble is not None and
               result.ensemble.volume is not None):
                nframes = observable_dict['Pressure'].size
                observable_dict['Volume'] = np.ones(nframes) * result.ensemble.volume

            result.observables = ObservableData()
            for key, gmxkey in self.__gmx_energy_names.items():
                result.observables[key] = observable_dict[gmxkey]

        return result



###########################################################################
#                                                                         #
#    physical_validation,                                                 #
#    a python package to test the physical validity of MD results         #
#                                                                         #
#    Written by Michael R. Shirts <michael.shirts@colorado.edu>           #
#               Pascal T. Merz <pascal.merz@colorado.edu>                 #
#                                                                         #
#    Copyright (C) 2012 University of Virginia                            #
#              (C) 2017 University of Colorado Boulder                    #
#                                                                         #
#    This library is free software; you can redistribute it and/or        #
#    modify it under the terms of the GNU Lesser General Public           #
#    License as published by the Free Software Foundation; either         #
#    version 2.1 of the License, or (at your option) any later version.   #
#                                                                         #
#    This library is distributed in the hope that it will be useful,      #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of       #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU    #
#    Lesser General Public License for more details.                      #
#                                                                         #
#    You should have received a copy of the GNU Lesser General Public     #
#    License along with this library; if not, write to the                #
#    Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,     #
#    Boston, MA 02110-1301 USA                                            #
#                                                                         #
###########################################################################
r"""
GROMACS python interface.

.. warning:: This is a mere place holder, as an official python API is
   currently being developed by the gromacs development team. It is
   probably neither especially elegant nor especially safe. Use of this
   module in any remotely critical application is strongly discouraged.
"""
import errno
import os
import sys
import subprocess
import re
import numpy as np


class GromacsInterface(object):
    def __init__(self, exe=None, dp=None, includepath=None):

        self._exe = None
        self._dp = False
        self._includepath = None

        if dp is not None:
            self.dp = dp

        if exe is None:
            # check whether 'gmx' / 'gmx_d' is in the path
            if self._check_exe(quiet=True, exe='gmx'):
                self.exe = 'gmx'
            elif self._check_exe(quiet=True, exe='gmx_d'):
                self.exe = 'gmx_d'
            else:
                print('WARNING: gmx executable not found. Set before attempting to run!')
        else:
            self.exe = exe

        if includepath is not None:
            self.includepath = includepath

    @property
    def exe(self):
        """exe is a string pointing to the gmx executable."""
        return self._exe

    @exe.setter
    def exe(self, exe):
        if self._check_exe(exe=exe):
            if os.path.dirname(exe):
                exe = os.path.abspath(exe)
            self._exe = exe

    @property
    def double(self):
        """double is a bool defining whether the simulation was ran at double precision"""
        return self._dp

    @double.setter
    def double(self, dp):
        assert isinstance(dp, bool)
        self._dp = dp

    @property
    def includepath(self):
        """includepath defines a path the parser looks for system files"""
        return self._includepath

    @includepath.setter
    def includepath(self, path):
        if isinstance(path, str):
            path = [path]
        self._includepath = path

    def get_quantities(self, edr, quantities, cwd=None,
                       begin=None, end=None, args=None):

        if args is None:
            args = []

        tmp_xvg = 'gmxpy_' + os.path.basename(edr).replace('.edr', '') + '.xvg'
        if cwd is not None:
            tmp_xvg = os.path.join(cwd, tmp_xvg)

        q_dict = {}

        for q in quantities:
            not_found = self._create_xvg(edr, tmp_xvg, [q], cwd=cwd,
                                         begin=begin, end=end, args=args)[1]
            if q in not_found:
                q_dict[q] = None
                continue

            skip_line = re.compile("^[#,@]")
            values = []
            times = []
            with open(tmp_xvg, 'r') as xvg:
                for line in xvg:
                    if skip_line.match(line):
                        continue
                    times.append(float(line.split()[0]))
                    values.append(float(line.split()[1]))

            if 'time' in q_dict:
                if not np.array_equal(np.array(times), q_dict['time']):
                    print('WARNING: Time discrepancy in ' + edr)
            else:
                q_dict['time'] = np.array(times)
            q_dict[q] = np.array(values)

            os.remove(tmp_xvg)

        return q_dict

    def read_trr(self, trr):
        tmp_dump = 'gmxpy_' + os.path.basename(trr).replace('.trr', '') + '.dump'
        with open(tmp_dump, 'w') as dump_file:
            proc = self._run('dump', ['-f', trr], stdout=dump_file, stderr=subprocess.PIPE)
            proc.wait()

        position = []
        velocity = []
        force = []
        box = []
        with open(tmp_dump) as dump:
            x = []
            v = []
            f = []
            b = []
            for line in dump:
                if 'frame' in line:
                    # new frame
                    if len(x) > 0:
                        # not the first frame - nothing to save there
                        position.append(np.array(x))
                        velocity.append(np.array(v))
                        force.append(np.array(f))
                        box.append(np.array(b))
                    x = []
                    v = []
                    f = []
                    b = []
                    continue
                if 'x[' in line:
                    x.append([float(l.strip()) for l in
                              line.split('{', 1)[1].split('}')[0].split(',')])
                if 'v[' in line:
                    v.append([float(l.strip()) for l in
                              line.split('{', 1)[1].split('}')[0].split(',')])
                if 'f[' in line:
                    f.append([float(l.strip()) for l in
                              line.split('{', 1)[1].split('}')[0].split(',')])
                if 'box[' in line:
                    b.append([float(l.strip()) for l in
                              line.split('{', 1)[1].split('}')[0].split(',')])
            # end loop over file - save last arrays
            position.append(np.array(x))
            velocity.append(np.array(v))
            force.append(np.array(f))
            box.append(np.array(b))

        result = {}
        for key, vector in zip(['position', 'velocity', 'force', 'box'],
                               [position, velocity, force, box]):
            vector = np.array(vector)
            if vector.size > 0:
                result[key] = vector
            else:
                result[key] = None
        return result

    @staticmethod
    def read_gro(gro):
        with open(gro) as conf:
            x = []
            v = []
            b = []
            title = conf.readline()
            while title:
                natoms = int(conf.readline().strip())
                for _ in range(natoms):
                    line = conf.readline()[20:]
                    line = line.split()
                    x.append([float(xx) for xx in line[0:3]])
                    v.append([float(vv) for vv in line[3:6]])

                line = conf.readline()
                line = line.split()
                b.append([float(vv) for vv in line[0:3]])
                title = conf.readline()

        result = {}
        for key, vector in zip(['position', 'velocity', 'force', 'box'],
                               [x, v, [], b]):
            vector = np.array(vector)
            if vector.size > 0:
                result[key] = vector
            else:
                result[key] = None
        return result

    @staticmethod
    def read_mdp(mdp):
        result = {}
        with open(mdp) as f:
            for line in f:
                line = line.split(';')[0].strip()
                if not line:
                    continue
                line = line.split('=')
                # unify mdp options - all lower case, only dashes
                option = line[0].strip().replace('_', '-').lower()
                if option not in ['include', 'define']:
                    value = line[1].strip().replace('_', '-').lower()
                else:
                    value = line[1].strip()
                result[option] = value
        return result

    @staticmethod
    def write_mdp(options, mdp):
        with open(mdp, 'w') as f:
            for key, value in options.items():
                f.write('{:24s} = {:s}\n'.format(key, value))

    def read_system_from_top(self, top, define=None, include=None):
        if not define:
            define = []
        else:
            define = [d.strip() for d in define.split('-D') if d.strip()]
        if not include:
            include = [os.getcwd()]
        else:
            include = [os.getcwd()] + [i.strip() for i in include.split('-I') if i.strip()]
        superblock = None
        block = None
        nmoleculetypes = 0
        topology = {}
        with open(top) as f:
            content = self._read_top(f, include=include, define=define)

        for line in content:
            if line[0] == '[' and line[-1] == ']':
                block = line.strip('[').strip(']').strip()
                if block == 'defaults' or block == 'system':
                    superblock = block
                    topology[superblock] = {}
                if block == 'moleculetype' or block == 'molecule_type':
                    nmoleculetypes += 1
                    superblock = block + '_' + str(nmoleculetypes)
                    topology[superblock] = {}
                continue
            if superblock is None or block is None:
                raise IOError('Not a valid .top file.')
            if block in topology[superblock]:
                topology[superblock][block].append(line)
            else:
                topology[superblock][block] = [line]

        for n in range(1, nmoleculetypes + 1):
            superblock = 'moleculetype_' + str(n)
            molecule = topology[superblock]['moleculetype'][0].split()[0]
            topology[molecule] = topology.pop(superblock)

        atomtype_list = topology['defaults']['atomtypes']
        topology['defaults']['atomtypes'] = {}
        for atomtype in atomtype_list:
            code = atomtype.split()[0]
            topology['defaults']['atomtypes'][code] = atomtype

        molecules = []
        for line in topology['system']['molecules']:
            molecule = line.split()[0]
            nmolecs = int(line.split()[1])
            natoms = len(topology[molecule]['atoms'])

            masses = []
            for atom in topology[molecule]['atoms']:
                if len(atom.split()) >= 8:
                    masses.append(float(atom.split()[7]))
                else:
                    code = atom.split()[1]
                    masses.append(float(topology['defaults']['atomtypes'][code].split()[3]))

            nbonds = 0
            nbondsh = 0
            bonds = []
            bondsh = []
            if 'bonds' in topology[molecule]:
                for bond in topology[molecule]['bonds']:
                    bond = bond.split()
                    a1 = int(bond[0]) - 1
                    a2 = int(bond[1]) - 1
                    m1 = masses[a1]
                    m2 = masses[a2]
                    if m1 > 1.008 and m2 > 1.008:
                        nbonds += 1
                        bonds.append([a1, a2])
                    else:
                        nbondsh += 1
                        bondsh.append([a1, a2])

            nangles = 0
            nanglesh = 0
            angles = []
            anglesh = []
            if 'angles' in topology[molecule]:
                for angle in topology[molecule]['angles']:
                    angle = angle.split()
                    a1 = int(angle[0]) - 1
                    a2 = int(angle[1]) - 1
                    a3 = int(angle[2]) - 1
                    m1 = masses[a1]
                    m2 = masses[a2]
                    m3 = masses[a3]
                    if m1 > 1.008 and m2 > 1.008 and m3 > 1.008:
                        nangles += 1
                        angles.append([a1, a2, a3])
                    else:
                        nanglesh += 1
                        anglesh.append([a1, a2, a3])

            settle = False
            if 'settles' in topology[molecule]:
                settle = True
                bonds = []
                bondsh = [[0, 1],
                          [0, 2],
                          [1, 2]]

            molecules.append({
                'name': molecule,
                'nmolecs': nmolecs,
                'natoms': natoms,
                'mass': masses,
                'nbonds': [nbonds, nbondsh],
                'bonds': bonds,
                'bondsh': bondsh,
                'nangles': [nangles, nanglesh],
                'angles': angles,
                'anglesh': anglesh,
                'settles': settle
            })

        return molecules

    def grompp(self, mdp, top, gro, tpr=None,
               cwd='.', args=None,
               stdin=None, stdout=None, stderr=None):
        cwd = os.path.abspath(cwd)
        assert os.path.exists(os.path.join(cwd, mdp))
        assert os.path.exists(os.path.join(cwd, top))
        assert os.path.exists(os.path.join(cwd, gro))

        if args is None:
            args = []

        if tpr is None:
            tpr = os.path.basename(mdp).replace('.mdp', '') + '.tpr'
        else:
            assert os.path.exists(os.path.join(cwd, os.path.dirname(tpr)))

        args = ['-f', mdp, '-p', top, '-c', gro, '-o', tpr] + args
        proc = self._run('grompp', args, cwd=cwd,
                         stdin=stdin, stdout=stdout, stderr=stderr)
        proc.wait()
        return proc.returncode

    def mdrun(self, tpr, edr=None, deffnm=None, cwd='.', args=None,
              stdin=None, stdout=None, stderr=None, mpicmd=None):
        cwd = os.path.abspath(cwd)
        tpr = os.path.join(cwd, tpr)
        assert os.path.exists(cwd)
        assert os.path.exists(tpr)

        if args is None:
            args = []

        if deffnm is None:
            deffnm = os.path.basename(tpr).replace('.tpr', '')

        args = ['-s', tpr, '-deffnm', deffnm] + args
        if edr is not None:
            args += ['-e', edr]
        proc = self._run('mdrun', args, cwd=cwd,
                         stdin=stdin, stdout=stdout, stderr=stderr,
                         mpicmd=mpicmd)
        proc.wait()
        return proc.returncode

    def _check_exe(self, quiet=False, exe=None):
        if exe is None:
            exe = self._exe
        try:
            exe_out = subprocess.run([exe, '--version'],
                                     stdout=subprocess.PIPE,
                                     universal_newlines=True).stdout
        except OSError as e:
            if hasattr(errno, 'ENOENT') and e.errno == errno.ENOENT:
                # file not found error.
                if not quiet:
                    print('ERROR: gmx executable not found')
                    print(exe)
                return False
            else:
                raise e
        # check that output is as expected
        return re.search(r':-\) GROMACS - gmx.* \(-:', exe_out)

    def _run(self, cmd, args, cwd=None, stdin=None, stdout=None, stderr=None, mpicmd=None):
        if self.exe is None:
            raise RuntimeError('Tried to use GromacsParser before setting gmx executable.')
        if mpicmd:
            command = [mpicmd, self.exe, cmd]
        else:
            command = [self.exe, cmd]
        command.extend(args)
        return subprocess.Popen(command, cwd=cwd,
                                stdin=stdin, stdout=stdout, stderr=stderr)

    def _create_xvg(self, edr, xvg, quantities, cwd=None,
                    begin=None, end=None, args=None):
        assert os.path.exists(edr)
        assert os.path.exists(os.path.abspath(os.path.dirname(xvg)))

        if args is None:
            args = []

        if self._dp:
            args.append('-dp')
        if begin is not None:
            args.extend(['-b', str(begin)])
        if end is not None:
            args.extend(['-e', str(end)])

        quants = ''
        for q in quantities:
            quants += str(q) + '\n'

        args = ['-f', edr, '-o', xvg] + args
        proc = self._run('energy', args, cwd=cwd,
                         stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        err = proc.communicate(quants.encode(sys.stdin.encoding))[1]

        encoding = sys.stderr.encoding
        if encoding is None:
            encoding = 'UTF-8'

        not_found = []
        if 'does not match anything' in err.decode(encoding):
            for q in quantities:
                if "String '" + q + "' does not match anything" in err.decode(encoding):
                    not_found.append(q)

        return proc.wait(), not_found

    def _read_top(self, filehandler, include, define):
        read = [True]
        content = []
        include_dirs = include
        if self.includepath:
            include_dirs += self.includepath
        for line in filehandler:
            line = line.split(';')[0].strip()
            line = line.split('*')[0].strip()
            if not line:
                continue
            if line[0] == '#':
                if line.startswith('#ifdef'):
                    option = line.replace('#ifdef', '').strip()
                    if option in define:
                        read.append(True)
                    else:
                        read.append(False)
                elif line.startswith('#ifndef'):
                    option = line.replace('#ifndef', '').strip()
                    if option not in define:
                        read.append(True)
                    else:
                        read.append(False)
                elif line.startswith('#else'):
                    read[-1] = not read[-1]
                elif line.startswith('#endif'):
                    read.pop()
                elif line.startswith('#define') and all(read):
                    option = line.replace('#define', '').strip()
                    define.append(option)
                elif line.startswith('#include') and all(read):
                    filename = line.replace('#include', '').strip().replace('"', '').replace('\'', '')
                    for idir in include_dirs:
                        try:
                            ifile = open(os.path.join(idir, filename))
                            break
                        except FileNotFoundError:
                            pass
                    else:
                        msg = ('Include file in .top file not found: ' +
                               line + '\n' +
                               'Include directories: ' + str(include_dirs))
                        raise IOError(msg)
                    if ifile:
                        subcontent = self._read_top(ifile,
                                                    [os.path.dirname(ifile.name)] + include,
                                                    define)
                        content.extend(subcontent)
                elif all(read):
                    raise IOError('Unknown preprocessor directive in .top file: ' +
                                  line)
                continue
            # end line starts with '#'

            if all(read):
                content.append(line)

        return content



###########################################################################
#                                                                         #
#    physical_validation,                                                 #
#    a python package to test the physical validity of MD results         #
#                                                                         #
#    Written by Michael R. Shirts <michael.shirts@colorado.edu>           #
#               Pascal T. Merz <pascal.merz@colorado.edu>                 #
#                                                                         #
#    Copyright (C) 2012 University of Virginia                            #
#              (C) 2017 University of Colorado Boulder                    #
#                                                                         #
#    This library is free software; you can redistribute it and/or        #
#    modify it under the terms of the GNU Lesser General Public           #
#    License as published by the Free Software Foundation; either         #
#    version 2.1 of the License, or (at your option) any later version.   #
#                                                                         #
#    This library is distributed in the hope that it will be useful,      #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of       #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU    #
#    Lesser General Public License for more details.                      #
#                                                                         #
#    You should have received a copy of the GNU Lesser General Public     #
#    License along with this library; if not, write to the                #
#    Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,     #
#    Boston, MA 02110-1301 USA                                            #
#                                                                         #
###########################################################################

import numpy as np
import warnings


def plot(res, legend=None, title=None,
         xlabel=None, ylabel=None, xlim=None, ylim=None,
         inv_x=False, inv_y=False, sci_x=False, sci_y=False,
         axtext=None, annotation_location=None,
         filename=None, screen=True):

    try:
        import matplotlib as mpl
        import matplotlib.pyplot as plt
        from matplotlib.ticker import AutoMinorLocator
    except ImportError:
        warnings.warn('Install matplotlib to enable plotting.')
        return

    font = {'family': 'serif',
            'weight': 'normal',
            'size': 16}
    mpl.rc('font', **font)

    plt.ioff()
    fig, ax = plt.subplots()
    xmin = float('inf')
    xmax = float('-inf')
    for r in res:
        x = r['x']
        y = r['y']
        if xlim is not None:
            x = x[(r['x'] >= xlim[0]) & (r['x'] <= xlim[1])]
            y = y[(r['x'] >= xlim[0]) & (r['x'] <= xlim[1])]
        if 'y_err' in r:
            dy = r['y_err']
            if xlim is not None:
                dy = dy[(r['x'] >= xlim[0]) & (r['x'] <= xlim[1])]
            ax.errorbar(x, y, label=r['name'], yerr=dy)
        else:
            ax.plot(x, y, label=r['name'])
        if xlim is not None:
            xmin = min(np.min(r['x']), xmin)
            xmax = max(np.max(r['x']), xmax)
        else:
            xmin = min(np.min(r['x']), xmin)
            xmax = max(np.max(r['x']), xmax)

    if legend is not None:
        ax.legend(loc=legend)
    box = ax.get_position()
    if title is not None:
        ax.set_title(title, y=1.05)
        box = box.from_bounds(box.x0, box.y0, box.width, box.height * 0.95)
    if xlabel is not None:
        ax.set_xlabel(xlabel, labelpad=5)
        box = box.from_bounds(box.x0, box.y0 + 0.05 * box.height, box.width, box.height * 0.95)
    if ylabel is not None:
        ax.set_ylabel(ylabel, labelpad=10)
        box = box.from_bounds(box.x0 + 0.05 * box.width, box.y0, box.width * 0.95, box.height)
    ax.set_position([box.x0, box.y0, box.width, box.height])
    ax.axis('auto')
    if xlim is not None:
        ax.set_xlim(xlim)
    else:
        ax.set_xlim([xmin, xmax])
    if ylim is not None:
        ax.set_ylim(ylim)
    ax.xaxis.set_minor_locator(AutoMinorLocator(2))

    if inv_x:
        ax.invert_xaxis()
    if inv_y:
        ax.invert_yaxis()

    if axtext is not None:
        if isinstance(axtext, str):
            axtext = [axtext]
        if annotation_location is None:
            annotation_location = [None for _ in axtext]
        if isinstance(annotation_location, tuple):
            annotation_location = [annotation_location]
        for t, loc in zip(axtext, annotation_location):
            bbox = dict(boxstyle="round", fc="w", ec="0.5", alpha=0.9)
            if loc is None:
                ax.text(0.95, 0.05, t, transform=ax.transAxes,
                        ha='right', va='bottom',
                        bbox=bbox)
            else:
                ax.text(loc[0], loc[1], t,
                        bbox=bbox)

    if sci_x:
        ax.ticklabel_format(style='sci', axis='x', scilimits=(-3, 4))
    if sci_y:
        ax.ticklabel_format(style='sci', axis='y', scilimits=(-3, 4))
    ax.xaxis.major.formatter._useMathText = True

    if filename is not None:
        fig.savefig(filename + '.pdf', dpi=300)
    if screen:
        fig.show()



###########################################################################
#                                                                         #
#    physical_validation,                                                 #
#    a python package to test the physical validity of MD results         #
#                                                                         #
#    Written by Michael R. Shirts <michael.shirts@colorado.edu>           #
#               Pascal T. Merz <pascal.merz@colorado.edu>                 #
#                                                                         #
#    Copyright (C) 2012 University of Virginia                            #
#              (C) 2017 University of Colorado Boulder                    #
#                                                                         #
#    This library is free software; you can redistribute it and/or        #
#    modify it under the terms of the GNU Lesser General Public           #
#    License as published by the Free Software Foundation; either         #
#    version 2.1 of the License, or (at your option) any later version.   #
#                                                                         #
#    This library is distributed in the hope that it will be useful,      #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of       #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU    #
#    Lesser General Public License for more details.                      #
#                                                                         #
#    You should have received a copy of the GNU Lesser General Public     #
#    License along with this library; if not, write to the                #
#    Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,     #
#    Boston, MA 02110-1301 USA                                            #
#                                                                         #
###########################################################################
r"""
Module containing the custom exception classes for the physical_validation
package.
"""


class PhysicalValidationError(Exception):
    r"""Base class for exceptions in the physical_validation module."""
    pass


class InputError(PhysicalValidationError):
    r"""Exception raised for input errors"""

    def __init__(self, argument, message):
        r"""

        Parameters
        ----------
        argument : string or list of strings
        message : string
        """
        self.argument = argument
        self.message = message


class ParserValueNotSetError(PhysicalValidationError):
    r"""
    Exception raised if a requested data value
    was not set by the user previously
    """

    def __init__(self, message):
        r"""

        Parameters
        ----------
        message : string
        """
        self.message = message


class FileFormatError(PhysicalValidationError):
    r"""Exception raised for files not following expected format"""

    def __init__(self, argument, message):
        r"""

        Parameters
        ----------
        argument : string or list of strings
        message : string
        """
        self.argument = argument
        self.message = message
